/*
 * SD.h
 * SD card and file manipulation
 * Created: 2013/11/06 1:16:33
 *  Author: Koichi Nishida
 */ 
/*
Copyright (C) 2016 Koichi NISHIDA
email to Koichi NISHIDA: tulip-house@msf.biglobe.ne.jp

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SD_H_
#define SD_H_

#include <avr/io.h>

#define EJECT (/*PORTA_IN*/VPORT0_IN&PIN5_bm)
#define WP (/*PORTC_IN*/VPORT1_IN&PIN3_bm)
#define SPI_SS_bm	0b00010000
#define SPI_MISO_bm	0b01000000
#define SPI_MOSI_bm	0b00100000
#define SPI_SCK_bm	0b10000000
#define DISABLE_CS PORTC_OUTSET=SPI_SS_bm
#define ENABLE_CS PORTC_OUTCLR=SPI_SS_bm

// SD card properties
struct SD {
	uint8_t *buff;
	uint8_t *buff2;
	uint8_t inited;
	uint32_t bpbAddr;
	uint8_t blkAdrAccs;			// 1 if block address access
	uint16_t bytesPerSector;
	uint8_t sectorsPerCluster;
	uint32_t fatAddr;			// the beginning of FAT
	uint32_t fatSectors;
	uint32_t fatSz;
	uint8_t numFats;
	uint32_t rootAddr;			// the beginning of RDE
	uint32_t rootSectors;
	uint32_t userAddr;			// the beginning of user area
	uint8_t fat32;				// 0 : fat16, 1 : fat32
};
extern struct SD SD_p;

#define FAT_ELEMS 64

// FILE properties
struct FILE {
	uint8_t valid;
	uint32_t entryAddress;		// entry address
	uint16_t entryOffset;		// entry offset
	uint32_t startCluster;		// start cluster
	uint32_t fat[FAT_ELEMS];	// the fat for the file
	uint32_t prevFatNum;		// previous Fat number for the file
	uint32_t prevClstNum;		// previous cluster number for the file
	uint8_t protect;			// write protect
	uint32_t length;			// file length
	uint32_t isDir;
	char name[12];				// file name and extension
	uint8_t written;
	uint32_t fat_seq;			// for sequential access
	uint32_t sectCount;			//
	uint32_t seq_curpos;		//		current position
};

// used by UI
struct FILELST {
	uint32_t blkAdr;
	uint16_t offset;
};

// write a byte data to spi
void SPI_writeByte(uint8_t c, uint8_t *err);

// write bytes data to spi
void SPI_writeBytes(uint8_t *a, uint8_t n, uint8_t *err);

// read data from spi
uint8_t SPI_readByte(uint8_t *err);

// initialization SD card
void SD_init(uint8_t *err);

// change buffer
void SD_changeBuff(uint8_t *b);

// wait until SD initialized
// return 1 if newly detected
uint8_t SD_detect(uint8_t start);

// write byte data one by one to the SD card
void SD_writeBytes(uint32_t adr, uint16_t ofs, uint8_t *ptr, uint16_t length, uint8_t *err);

// write a word one by one to the SD card
void SD_writeWord(uint32_t adr, uint16_t ofs, uint16_t d, uint8_t *err);

// write a double word one by one to the SD card
void SD_writeDWord(uint32_t adr, uint16_t ofs, uint32_t d, uint8_t *err);

// begin read data from shared memory
void SD_readSharedMemoryBegin(uint32_t adr, uint16_t len, uint8_t *err);

// end read data from shared memory
void SD_readSharedMemoryEnd(uint8_t *err);

// begin write data to shared memory
void SD_writeSharedMemoryBegin(uint32_t adr, uint16_t len, uint8_t *err);

// end write data to shared memory
void SD_writeSharedMemoryEnd(uint8_t *err);

// open the file
void FILE_open(struct FILE *filep, int32_t dir_cluster, char *name, char *exts, uint8_t seeHidden, uint8_t *err);

// open the file with the absolute path
// name should be <8 character directory name><3 character extension><8...><3...>...0
void FILE_openAbs(struct FILE *filep, char *name, uint8_t seeHidden, uint8_t *err);

// read a sector from the block address
void SD_readBlock(uint32_t block_adr, uint8_t *err);

// read a sector from the file
void FILE_read(struct FILE *filep, uint32_t long_sector, uint8_t *err);

// prepare reading a sector from the file
void FILE_readBegin(struct FILE *filep, uint32_t long_sector, uint8_t *err);

// end the reading
void FILE_readEnd(uint8_t *err);

// prepare writing a sector to the file
void FILE_writeBegin(struct FILE *filep, uint32_t long_sector, uint8_t *err);

// end the writing
void FILE_writeEnd(uint8_t *err);

// get file name, extension, attribute and start cluster from an file list entry
// used in UI.c
void FILE_getEntry(struct FILELST *e, char *name, char *ext, uint8_t *attr, uint32_t *stclst, uint8_t *err);

// traverse a directory and search files whose extension is exts.
void FILE_traverseADirectory(int32_t dir_cluster, char *exts, struct FILELST *list, uint16_t *entryNum, void (*subFunc)(struct FILELST *list, uint16_t *entryNum, uint32_t blkAdr, uint16_t offset), uint8_t *err);

// make file list and sort
// used in UI.c
uint16_t FILE_makeList(int32_t dir_cluster, struct FILELST *list, char *targExt, uint8_t *err);

// create a file
void FILE_create(int32_t dir_cluster, const char *name, uint32_t length, uint8_t *err);

// create a file with the absolute path
// name should be <8 character directory name><3 character extension><8...><3...>...0
void FILE_createAbs(struct FILE *filep, char *name, uint32_t length, uint8_t *err);

// substitute the extension of the full path file name with the given new extension
void FILE_substituteFullpathExtwith(char *fullpath, char *ext);

// get a short file name from a long file name written in shared memory
void FILE_getShortFileName(uint8_t *short_name, uint8_t *err);

// read 4 characters from shared memory and convert them to a 4 byte word.
uint16_t FILE_readShared4(uint8_t *err);

// sequential block read from the file
void FILE_seq_read(struct FILE *fp, uint8_t *err);

// read a byte sequentially from the file
void FILE_seq_read_byte(struct FILE *fp, uint8_t *eof, uint8_t *bt, uint8_t *err);

// read a word sequentially from the file
void FILE_seq_read_word(struct FILE *fp, uint8_t *eof, uint16_t *wd, uint8_t *err);

// read a long word sequentially from the file
void FILE_seq_read_dword(struct FILE *fp, uint8_t *eof, uint32_t *dwd, uint8_t *err);

// write a byte sequentially to the file
void FILE_seq_write_byte(struct FILE *fp, uint8_t bt, uint8_t *err);

// write a word sequentially to the file
void FILE_seq_write_word(struct FILE *fp, uint16_t wd, uint8_t *err);

// write a dword sequentially to the file
void FILE_seq_write_dword(struct FILE *fp, uint32_t dwd, uint8_t *err);

// reset the sequential read position to the top of the file
void FILE_seq_reset(struct FILE *fp);

// seek from the current sequential read position
void FILE_seq_seek_bytes(struct FILE *fp, uint32_t length, uint8_t *err);

// end writing a byte sequentially to the file
void FILE_seq_write_byte_end(struct FILE *fp, uint8_t *err);

// rename the file
void FILE_rename(struct FILE *fp, char *name_ext,uint8_t *err);

#endif /* SD_H_ */